#!/usr/bin/perl

# Try to keep the programming style good ;-)

use strict;

# Commented out for compatibility reasons because older Perl
# distributions don't know the warnings pragma.
#use warnings;

# Some modules we need for password-related Stuff

use Crypt::PasswdMD5;
use Crypt::RandPasswd;

# We want to parse commandline arguments in a comfortable way

use Getopt::Std;

# We need this to determine the OS we run on.

use Config;

# Parse the commandline arguments and fall back to default values if
# neccessary

my %opts;
getopts('b:g:i:l:S:o:m:x:e:D:t:f:G:u:T:O::chrsdy', \%opts);

if ( $opts{h} ) 
	{
	if ( $0 =~ m/userquick/ ) { usage_userquick(); }
	if ( $0 =~ m/userneu/ ) { usage_userneu(); }
	}

sub usage_userneu{ die "Usage:
    userneu [ -cdhrsy ] [ -i input file ] [ -g [primary group] ]
    [ -b [base directory for \$HOME] ] [ -S [smbpasswd binary] ]
    [ -o [Output file] ] [ -m [minimum password length] ]
    [ -x [maximum password length] ] [ -D [descriptors] ]
    [ -f [configuration file] ] [ -G [member groups] ] [ -t [delimiter] ]
    [ [ -T [target system] ] [ -O [shellscript output] ]\n"; }

sub usage_userquick{ die "Usage:
    userquick [ -cdhrsy ] [ -u user information of single user ]
    [ -g primary group ] [ -b [base directory for \$HOME] ]
    [ -S [smbpasswd binary] ] [ -o [Output file] ]
    [ -m [minimum password length] ] [ -x [maximum password length] ]
    [ -D [descriptors] ] [ -f [configuration file] ] [ -G [member groups] ]
    [ [ -T [target system] ] [ -O [shellscript output] ]\n"; }


# Set the default values for some options. These are overridden by the
# settings in the configuration file which are in turn overridden by the
# options given on the command line.

my $userlist = "userlist";
my $group = "users";
my $homebase = "/home/$group";
my $smbpass = "/usr/local/bin/smbpasswd";
my $outfile = "useraccounts";
my $passtype = "letters";
my $descriptor_raw = "username;real name;class";
my $delimiter = '/\t/';
my $paranoid_pass;
my $minlength = 8;
my $maxlength = 8;
my $generate_samba;
my $schoolyear_add;
my $dupe_engine_enable;
my $errlog = "duplicate_users";
my $member_groups;
my $pass_minlength;
my $pass_maxlength;
my $create_shellscript;
my $target_system;
my $shellscript_output;

# Check right now if a different configuration file was specified on the
# command line as we parse the configuration file before looking at the
# other command line options

my $config = "/etc/userneu.conf";
if ( $opts{f} ) { $config = $opts{f}; }

# Read the configuration file

if ( -f $config ) { config_file_parse(); }
else { warn "No config file found, using command line options only\n" }

sub config_file_parse
{
{ package Configure; do $config; }


if ( $Configure::userlist ) { $userlist = $Configure::userlist; }
if ( $Configure::group ) { $group = $Configure::group; }
if ( $Configure::homebase ) { $homebase = $Configure::homebase; }
if ( $Configure::smbpass ) {$smbpass = $Configure::smbpass; }
if ( $Configure::outfile ) {$outfile = $Configure::outfile; }
if ( $Configure::passtype ) {$passtype = $Configure::passtype; }
if ( $Configure::descriptor ) {$descriptor_raw = $Configure::descriptor; }
if ( $Configure::delimiter ) {$delimiter = $Configure::delimiter; }
if ( $Configure::paranoid_pass ) {$paranoid_pass = $Configure::paranoid_pass; }
if ( $Configure::generate_samba ) {$generate_samba = $Configure::generate_samba; }
if ( $Configure::schoolyear_add ) {$schoolyear_add = $Configure::schoolyear_add; }
if ( $Configure::dupe_engine_enable ) {$dupe_engine_enable = $Configure::dupe_engine_enable; }
if ( $Configure::errlog ) {$errlog = $Configure::errlog; }
if ( $Configure::member_groups ) {$member_groups = $Configure::member_groups; }
if ( $Configure::pass_minlength ) {$pass_minlength = $Configure::pass_minlength; }
if ( $Configure::pass_maxlength ) {$pass_maxlength = $Configure::pass_maxlength; }
if ( $Configure::create_shellscript ) {$create_shellscript = $Configure::create_shellscript; }
if ( $Configure::target_system ) {$target_system = $Configure::target_system; }
if ( $Configure::shellscript_output ) {$target_system = $Configure::shellscript_output; }
}

# Now check the command line options and override the values from the
# configuration file with them.

if ( $opts{i} ) { $userlist = $opts{i}; }

if ( $opts{g} ) { $group = $opts{g}; }

# Determine if we should get group information from input

my $group_field;
# Determine if we should read the primary group from the input file
my $group_from_file;
if ( $group =~ /^%(\d+)$/ ) 
	{ 
	$group_field = $1 - 2;
	$group_from_file = 1;
	}

if ( $opts{b} ) { $homebase = $opts{b}; }

if ( $opts{S} ) { $smbpass = $opts{S}; }

if ( $opts{o} ) { $outfile = $opts{o}; }

if ( $opts{r} ) { $paranoid_pass = 1 }

if ( $paranoid_pass )
	{
	$minlength = 10;
	$maxlength = 10;
	$passtype = "chars";
	}

if ( $opts{D} ) { $descriptor_raw = $opts{D}; }

my @descriptor = split(/\;/, $descriptor_raw);

if ( $opts{t} ) { $delimiter = $opts{t}; }

if ( $delimiter !~ /^\/.*\/$/ ) { die "$delimiter is not a valid regular expression. A correct example would be '/\\t/'.\n"; }

if ( $opts{d} ) { $dupe_engine_enable = 1; }

if ( $opts{G} ) { $member_groups = $opts{G}; }

# Determine if we should read the list of member groups from the input file

my $member_groups_field;
my $mgroups_from_file;
if ( $member_groups =~ /^%(\d+)$/ )
	{
	$member_groups_field = $1 - 2; 
	$mgroups_from_file = 1;
	}

if ( $opts{y} ) { $schoolyear_add = 1; }

if ( $opts{T} ) { $target_system = $opts{T}; }
if ( $opts{O} ) { $shellscript_output = $opts{O}; }

# Make $create_shellscript implicit if $target_system or 
# $shellscript_output are true.

if ( $target_system or $shellscript_output ) { $create_shellscript = 1; }

if ( $opts{c} )
	{
	$create_shellscript = 1;
	if ( ! $target_system ) { $target_system = $Config{osname}; }
	if ( ! $shellscript_output ) { $shellscript_output = "createusers.sh"; }
	}

if ( ! $create_shellscript )
	{ 
	# Check if we have to be root
	if ( $> != 0 ) {die "You must be root to run this.\n";}

	# Check if we have to make sure the primary group exists
	if ( ! $group_from_file and ! getgrnam $group ) { die "Group $group doesn't exist. Please create it or specify an existing group.\n" }
	}



# Do some chceks that are only necessary if a shell script is to be
# output

if ( $create_shellscript )
	{
	if ( $target_system ne "linux" and $target_system ne "freebsd" ) { die "Currently only Linux and FreeBSD are supported.\n"; }
	if ( ($shellscript_output eq "-") and ($outfile eq "-") ) { die "Only one thing at a time can be output on STDOUT, either the shellscript or the list of user accounts.\n"; }
	}

# Compose the member groups argument to useradd(8) (or pw(8) on FreeBSD)
# only if there is a list of member groups

my $member_groups_arg;
my @member_groups_check;

if ( $member_groups  and ! $mgroups_from_file)
	{
	@member_groups_check = split(/,/, $member_groups);

	# Now check if all the member groups specified do exist.

	if ( ! $create_shellscript ) { membergroups_check(@member_groups_check); }
	$member_groups_arg = "-G" . $member_groups;
	}



# Override the values for minimum and maximum password length if the -m
# or -x options are given

if ( $opts{m} ) { $minlength = $opts{m}; }
elsif ( $pass_minlength ) { $minlength = $pass_minlength; }
if ( $opts{x} ) { $maxlength = $opts{x}; }
elsif ( $pass_maxlength ) { $maxlength = $pass_maxlength; }

# Although we require the person running the script to supply the
# enclosing slashes of the regexp we remove them now as the pattern
# won't work otherwise.

$delimiter =~ s/^\/(.*)\/$/$1/;

warn "Group: $group\n";
warn "Base directory: $homebase\n";
warn "Userliste: $userlist\n\n";

# Determine the system we run on

my $os = $Config{osname};



# Create the filehandle to read the userlist from or read from STDIN if
# $opts{i} is "-"

if ( $userlist eq "-" ) { open LISTE,"&STDIN"; }
else
	{
	open LISTE,$userlist or die "$userlist: No such file or directory.\n";
	}

# Create the filehandle for the shellscript to be written if
# $create_shellscript is true.

if ( $create_shellscript )
	{
	if ( $shellscript_output eq "-" ) { open SHELLSCRIPT,"&STDOUT"; }
	else { open SHELLSCRIPT,">>$shellscript_output"; }
	print SHELLSCRIPT "#!/bin/sh\n";
	}

# Open the output filehandle

if ( $outfile eq "-" ) { open OUTFILE,"&STDOUT"; }
else { open OUTFILE,">>$outfile"; }



# Do some deep magic to determine the school year.

# First get the raw time
my @time = localtime;

# Extract the relevant parts

my $month = $time[4];
my $year = $time[5];

# Convert the year to 2-digit format

my $year1 = sprintf("%02d", $year % 100);

# Now it's time for the arcane part...

my $schoolyear;
if ($month le 7) {
	my $year2=$year1 - 1;
	$schoolyear = "0$year2/$year1";
}
else {
	my $year2=$year1 + 1;
	$schoolyear = "$year1/0$year2";
}

# Make s implicit if -S is given

if ( $opts{S} and ! $generate_samba )
	{
	$generate_samba = 1;
	}

# Check for smbpasswd(8) if $generate_samba is true

if ( $opts{s} ) 
	{
	$generate_samba = 1;
	if ( ! $create_shellscript and ! -x $smbpass ) { die "$smbpass doesn't exist or isn't executable. Please specify.\n"; }
	}

# Determine if we shall parse a list or create just one single account
# specified on the command line.

if ( $0 =~ /userneu/ ) { parse_file(); }
elsif ( $0 =~ /userquick/ ) { single_user(); }

# Create the variable in which the number of created user accounts will
# be stored.

my $usercount;

# Extract the account data from the list of usernames

sub parse_file
{
while (my $line = <LISTE>)
	{
	# Parse the userlist and put it's elements in variables

	my @entry = split(/$delimiter/, $line);

	# First check if @descriptor and @entry contain an equal number
	# of entries

	if ( (scalar @descriptor) != (scalar @entry) ) { die "The number of fields in the input file doesn't match the number of descriptors\n"; }

	my $name = shift @entry;
	my $output_rest;

	# Remove the newline left over from parsing the input file
	chomp @entry;

	# Extract the group from @entry

	if ( $group_from_file ) 
		{ 
		$group = $entry[$group_field]; 
		if ( ! $create_shellscript and ! getgrnam $group ) { die "Group $group doesn't exist. Please create it or specify an existing group.\n" }
		}

	my $membergroups;
	if ( $mgroups_from_file )
		{
		$membergroups = $entry[$member_groups_field];
		my @member_groups = split(/,/, $membergroups);
		member_groups_check(@member_groups);
		$member_groups_arg = "-G" . $entry[$member_groups_field];
		undef $entry[$member_groups_field];
		}
	
	# We put this in an enclosing block so we don't shift @entry and
	# @descriptor to pieces because we need them again later on.

		{
		my @descriptor = @descriptor;
		my @entry = @entry;

		# shorten @descriptor by one to match @entry
		shift @descriptor;
		
		while ( @entry && @descriptor)
			{
			$output_rest .= (shift @descriptor) . ": " . (shift @entry ). " ";
			}
		}
	
	# Determine if we should try appending random characters to duplicate
	# user names or simply log and report them. Only do this if
	# $create_shellscript is false.

	if ( ! $create_shellscript )
		{
		if ( $dupe_engine_enable ) { $name = dupe_engine($name,$output_rest); }
		else { next if dupe_log($name,$output_rest); }
		}

	# Compose the GECOS field for /etc/passwd

	my $gecos;
	while ( @entry )
		{
		$gecos .= (shift @entry) . ";";
		}

	if ( $schoolyear_add ) { $gecos .= $schoolyear };

	$gecos =~ s/\;$//;

	# Generate a password for the user

	my $clearpass = Crypt::RandPasswd->$passtype( $minlength, $maxlength );
	# Make sure the password contains no single qutoes (which cause
	# problems) if $paranoid_pass is true.
	if ( $paranoid_pass and ($clearpass =~ /'/) ) { until ( ! $clearpass =~ /'/ ) { $clearpass = Crypt::RandPasswd->$passtype( $minlength, $maxlength ); } }

	my $passwdhash = unix_md5_crypt ($clearpass, '$1$');

	# Now call the function to create the account or the function to
	# write the shell script.

	if ( !$create_shellscript ) { account_create($name,$clearpass,$passwdhash,$gecos,$member_groups_arg); }
	else { shellscript_create($name,$clearpass,$passwdhash,$gecos,$target_system,$member_groups_arg); }
	$usercount++;

	# Write the account data to a file

	print OUTFILE "Username: $name password: $clearpass $output_rest\n";

	# Only output the account information if OUTFILE doesn't point
	# to STDOUT as outputting it twice would be rather pointless.

	if ( $outfile ne "-" ) { print "Username: $name password: $clearpass $output_rest\n"; }
	
}


close LISTE;
close OUTFILE;

if ( $outfile ne "-" ) { chmod (0600, $outfile); }
warn "Successfully created $usercount users.\n";
}





sub single_user
{
if ( ! $opts{u} ) { die "Please specify the information of the user you want to be created.\n"; }
my @person = split(/;/, $opts{u});

if ( (scalar @person) != (scalar @descriptor) ) { die "The number of fields in the argument to -u doesn't match the number of descriptors\n"; }

my $name = shift @person;

# Shorten @descriptor by one to match @person
shift @descriptor;

my $output_rest;

	{

	my @person = @person;
	my @descriptor = @descriptor;

	# Compose the information that comes after the username in the output
	# that is printed  when the script is done.

	while ( @person && @descriptor)
	        {
	        $output_rest .= (shift @descriptor) . ": " . (shift @person) . " ";
		}
	}

# Determine if we should try appending random characters to duplicate
# user names or simply log and report them.

if ( $dupe_engine_enable ) { $name = dupe_engine($name); }
else { die "User $name already in password database.\n" if
dupe_log($name,$output_rest); }

# Compose the GECOS field for /etc/passwd

my $gecos;
while ( @person ) { $gecos .= (shift @person) . ";"; }

# Add the school year to the GECOS field if desiered

if ( $schoolyear_add ) { $gecos .= $schoolyear };

# Remove possibly leftover trailing semicolons from $gecos because it
# looks nicer that way ;-)

$gecos =~ s/\;$//;


# Generate a password for the user

my $clearpass = Crypt::RandPasswd->$passtype( $minlength, $maxlength );

my $passwdhash = unix_md5_crypt ($clearpass, '$1$');

# Now create the actual account or write the script if
# $create_shellscript is true.

if ( !$create_shellscript ) { account_create($name,$clearpass,$passwdhash,$gecos); }
else { shellscript_create($name,$clearpass,$passwdhash,$gecos,$target_system); }

print "Username: $name password: $clearpass $output_rest\n";
}





sub dupe_engine
	{
	# Check if the user exists
	my $username = $_[0];
	my $output_rest = $_[1];
	if ( getpwnam $username ) 
		{
		until ( ! getpwnam $username )
			{
			my $randomizer = Crypt::RandPasswd->word(1,2);
			$username .= $randomizer;
			warn "New Name: $username\n";
			}
		}
	return $username;
	}






sub dupe_log
	{
	# Check if the user exists
	my $username = $_[0];
	my $output_rest = $_[1];
	if ( getpwnam $username )
		{
		if ( $opts{e} ) { $errlog = $opts{e}; }
		open DUPELOG,">> $errlog";
		print DUPELOG "Username: $username $output_rest\n";
		close DUPELOG;
		}
		else {return;}
	}






sub account_create
	{
	# Create an Unix account
	# First determine which sytem we run on and which commands we
	# have to use

	my $username = $_[0];
	my $clearpass = $_[1];
	my $passwdhash = $_[2];
	my $gecos = $_[3];
	my $member_groups_arg = $_[4];

	if ( $os eq "linux" )
		{
		# useradd(8) is braindead because it won't let me
		# specify a base directory for $HOME
		my $home = "$homebase/$username";
		if ( $home =~ /\/{2,}/ ) { $home =~ s/\/{2,}/\//g; }

		my @useradd_cmd = ("/usr/sbin/useradd", "-m", "-c$gecos", "-g$group", "-d$home");
		if ( $member_groups_arg ) { push @useradd_cmd, $member_groups_arg;}
		push @useradd_cmd, $username;
		system @useradd_cmd;
		open PASSPIPE,"| /usr/sbin/chpasswd -e";
		print PASSPIPE "$username:$passwdhash\n";
		close PASSPIPE;
		}

	if ( $os eq "freebsd" )
		{
		my @pw_args = ("useradd", "-n $username", "-m", "-c \'$gecos\'", "-g $group", "-b $homebase", "-h 0");
		if ( $member_groups_arg ) { push @pw_args, $member_groups_arg; }
		open PWPIPE, "| /usr/sbin/pw @pw_args";
		print PWPIPE "$clearpass\n";
		close PWPIPE;
		}

		# Check $generate_samba and call smb_create() if true.
		# If this function does not have to be called on your
		# systems you are after all a rather lucky person.

		if ( $generate_samba ) { smb_create($username,$clearpass); }
		}




sub shellscript_create
	{
	# Create an Unix account
	# First determine which sytem we run on and which commands we
	# have to use

	my $username = $_[0];
	my $clearpass = $_[1];
	my $passwdhash = $_[2];
	my $gecos = $_[3];
	my $target_system = $_[4];
	my $member_groups_arg = $_[5];

	if ( $target_system eq "linux" )
		{
		# useradd(8) is braindead because it won't let me
		# specify a base directory for $HOME
		my $home = "$homebase/$username";
		if ( $home =~ /\/{2,}/ ) { $home =~ s/\/{2,}/\//g; }

		my @useradd_cmd = ("/usr/sbin/useradd", "-m", "-c$gecos", "-g$group", "-d$home");
		if ( $member_groups_arg ) { push @useradd_cmd, $member_groups_arg;}
		push @useradd_cmd, $username;
		print SHELLSCRIPT "@useradd_cmd\n";
		print SHELLSCRIPT "echo '$username:$passwdhash' | /usr/sbin chpasswd -e\n";
		}

	if ( $target_system eq "freebsd" )
		{
		my @pw_args = ("useradd", "-n $username", "-m", "-c \'$gecos\'", "-g $group", "-b $homebase", "-h 0");
		if ( $member_groups_arg ) { push @pw_args, $member_groups_arg; }
		print SHELLSCRIPT "echo '$clearpass' | /usr/sbin/pw @pw_args\n"
		}

		# Check $generate_samba and call smb_create() if true.
		# If this function does not have to be called on your
		# systems you are after all a rather lucky person.

		if ( $generate_samba ) { print SHELLSCRIPT "(echo '$clearpass';echo '$clearpass') | $smbpass -s -a $username\n" ; }
		}





sub membergroups_check
{
@member_groups_check = @_;
my @failed_groups;
foreach my $membergroup ( @member_groups_check )
	{
	# We need this workarround for existing groups with GID 0
	# (such as wheel on FreeBSD) because otherwise perl would
	# interpret that zero as false, causing the script to die.

	if ( ! getgrnam $membergroup )
		{
		warn "$membergroup doesn't exist\n";
		push @failed_groups, $membergroup;
		}
	}

# getgrnam returns 0 as group ID of wheel or similiar groups with GID
# zero (of course). Because perl would interpret this as false we need
# this little workarround so the script doesn't die if users are members
# of the wheel group.

if ( @failed_groups )  
		{
		foreach my $checkgroup ( @failed_groups )
			{
			if ( (getgrgid 0) eq $checkgroup ) { shift @failed_groups; }
			}
			if ( @failed_groups) {
			die "The following groups you specified as member groups for the users to be created do not exist: @failed_groups\n" }
		}
}
	



sub smb_create 
	{
	# Use smbpasswd(8) to create an smbpasswd entry
	# Yes this is ugly...and so is smbpasswd's non-interactive
	# mode...

	my $username = $_[0];
	my $clearpass = $_[1];
	my $smbpasswd = "$smbpass -s -a $username";
	open SMBPIPE,"| $smbpasswd";
	print SMBPIPE "$clearpass\n$clearpass\n";
	close SMBPIPE;
	}
